/* Generated by the protocol buffer compiler.  DO NOT EDIT! */
/* Generated from: ccx_sub_entry_message.proto */

/* Do not generate deprecated warnings for self */
#ifndef PROTOBUF_C__NO_DEPRECATED
#define PROTOBUF_C__NO_DEPRECATED
#endif

#include "ccx_sub_entry_message.pb-c.h"

#ifdef ENABLE_SHARING

void   ccx_sub_entry_message__init
(CcxSubEntryMessage         *message)
{
	static CcxSubEntryMessage init_value = CCX_SUB_ENTRY_MESSAGE__INIT;
	*message = init_value;
}
size_t ccx_sub_entry_message__get_packed_size
(const CcxSubEntryMessage *message)
{
	assert(message->base.descriptor == &ccx_sub_entry_message__descriptor);
	return protobuf_c_message_get_packed_size((const ProtobufCMessage*)(message));
}
size_t ccx_sub_entry_message__pack
(const CcxSubEntryMessage *message,
uint8_t       *out)
{
	assert(message->base.descriptor == &ccx_sub_entry_message__descriptor);
	return protobuf_c_message_pack((const ProtobufCMessage*)message, out);
}
size_t ccx_sub_entry_message__pack_to_buffer
(const CcxSubEntryMessage *message,
ProtobufCBuffer *buffer)
{
	assert(message->base.descriptor == &ccx_sub_entry_message__descriptor);
	return protobuf_c_message_pack_to_buffer((const ProtobufCMessage*)message, buffer);
}
CcxSubEntryMessage *
ccx_sub_entry_message__unpack
(ProtobufCAllocator  *allocator,
size_t               len,
const uint8_t       *data)
{
	return (CcxSubEntryMessage *)
		protobuf_c_message_unpack(&ccx_sub_entry_message__descriptor,
		allocator, len, data);
}
void   ccx_sub_entry_message__free_unpacked
(CcxSubEntryMessage *message,
ProtobufCAllocator *allocator)
{
	assert(message->base.descriptor == &ccx_sub_entry_message__descriptor);
	protobuf_c_message_free_unpacked((ProtobufCMessage*)message, allocator);
}
static const ProtobufCFieldDescriptor ccx_sub_entry_message__field_descriptors[6] =
{
	{
		"eos",
		1,
		PROTOBUF_C_LABEL_REQUIRED,
		PROTOBUF_C_TYPE_INT32,
		0,   /* quantifier_offset */
		offsetof(CcxSubEntryMessage, eos),
		NULL,
		NULL,
		0,             /* flags */
		0, NULL, NULL    /* reserved1,reserved2, etc */
	},
	{
		"stream_name",
		2,
		PROTOBUF_C_LABEL_REQUIRED,
		PROTOBUF_C_TYPE_STRING,
		0,   /* quantifier_offset */
		offsetof(CcxSubEntryMessage, stream_name),
		NULL,
		NULL,
		0,             /* flags */
		0, NULL, NULL    /* reserved1,reserved2, etc */
	},
	{
		"counter",
		3,
		PROTOBUF_C_LABEL_REQUIRED,
		PROTOBUF_C_TYPE_INT64,
		0,   /* quantifier_offset */
		offsetof(CcxSubEntryMessage, counter),
		NULL,
		NULL,
		0,             /* flags */
		0, NULL, NULL    /* reserved1,reserved2, etc */
	},
	{
		"start_time",
		4,
		PROTOBUF_C_LABEL_REQUIRED,
		PROTOBUF_C_TYPE_INT64,
		0,   /* quantifier_offset */
		offsetof(CcxSubEntryMessage, start_time),
		NULL,
		NULL,
		0,             /* flags */
		0, NULL, NULL    /* reserved1,reserved2, etc */
	},
	{
		"end_time",
		5,
		PROTOBUF_C_LABEL_REQUIRED,
		PROTOBUF_C_TYPE_INT64,
		0,   /* quantifier_offset */
		offsetof(CcxSubEntryMessage, end_time),
		NULL,
		NULL,
		0,             /* flags */
		0, NULL, NULL    /* reserved1,reserved2, etc */
	},
	{
		"lines",
		7,
		PROTOBUF_C_LABEL_REPEATED,
		PROTOBUF_C_TYPE_STRING,
		offsetof(CcxSubEntryMessage, n_lines),
		offsetof(CcxSubEntryMessage, lines),
		NULL,
		NULL,
		0,             /* flags */
		0, NULL, NULL    /* reserved1,reserved2, etc */
	},
};
static const unsigned ccx_sub_entry_message__field_indices_by_name[] = {
	2,   /* field[2] = counter */
	4,   /* field[4] = end_time */
	0,   /* field[0] = eos */
	5,   /* field[5] = lines */
	3,   /* field[3] = start_time */
	1,   /* field[1] = stream_name */
};
static const ProtobufCIntRange ccx_sub_entry_message__number_ranges[2 + 1] =
{
	{ 1, 0 },
	{ 7, 5 },
	{ 0, 6 }
};
const ProtobufCMessageDescriptor ccx_sub_entry_message__descriptor =
{
	PROTOBUF_C__MESSAGE_DESCRIPTOR_MAGIC,
	"ccx_sub_entry_message",
	"CcxSubEntryMessage",
	"CcxSubEntryMessage",
	"",
	sizeof(CcxSubEntryMessage),
	6,
	ccx_sub_entry_message__field_descriptors,
	ccx_sub_entry_message__field_indices_by_name,
	2, ccx_sub_entry_message__number_ranges,
	(ProtobufCMessageInit)ccx_sub_entry_message__init,
	NULL, NULL, NULL    /* reserved[123] */
};

#endif //ENABLE_SHARING
