message ccx_sub_entry_message {
    required int32 eos = 1;
    required string stream_name = 2;
    required int64 counter = 3;
    required int64 start_time = 4;
    required int64 end_time = 5;
    repeated string lines = 7;
}
